@@ -1,8 +1,8 @@
-{-# LANGUAGE DeriveDataTypeable, CPP #-}
+{-# LANGUAGE DeriveDataTypeable, CPP, DoAndIfThenElse #-}
 {- |
 
 shelltest - for testing command-line programs. See shelltestrunner.cabal.
-(c) Simon Michael 2009-2017, released under GNU GPLv3 or later.
+(c) Simon Michael & contributors 2009-2023, released under GNU GPLv3 or later.
 
 -}
 
@@ -11,6 +11,7 @@ where
 
 import Control.Concurrent (forkIO)
 import Control.Concurrent.MVar (newEmptyMVar, putMVar, takeMVar)
+import Control.Monad (void)
 import Data.Algorithm.Diff
 import System.Console.CmdArgs
 import System.Directory (doesDirectoryExist)
@@ -18,75 +19,87 @@ import System.FilePath (takeDirectory)
 import System.FilePath.Find (findWithHandler, (==?), always)
 import qualified System.FilePath.Find as Find (extension)
 import System.IO (Handle, hGetContents, hPutStr)
-import System.Process (StdStream (CreatePipe), shell, createProcess, CreateProcess (..), waitForProcess, ProcessHandle)
+import System.Process (StdStream (CreatePipe), shell, proc, createProcess, CreateProcess (..), waitForProcess, ProcessHandle)
 import Test.Framework (defaultMainWithArgs)
 import Test.Framework.Providers.HUnit (hUnitTestToTests)
 import Test.HUnit
 import Text.Parsec
+import Test.Hspec.Core.Runner (defaultConfig, runSpec, Config (..))
+import Test.Hspec.Contrib.HUnit (fromHUnitTest)
 
 import Import
 import Utils
 import Types
 import Parse
+import Print
 import Preprocessor
 
 
 progname, progversion :: String
 progname = "shelltest"
-progversion = progname ++ " " ++ "1.9.0.1"
+progversion = progname ++ " " ++ "1.10"
 proghelpsuffix :: [String]
 proghelpsuffix = [
    "shelltest file formats, tried in this order:"
   ,""
-  ,"Description                                            Delimiters, in order"
+  ,"Description                                            Delimiters, in order"    
   ,"------------------------------------------------------ ------------------------"
   ,"v2 input then multiple tests; some delimiters optional <<<    $$$ >>> >>>2 >>>="
-  ,"v3 same as v2, but with short delimiters               <      $   >   >2   >="
+  ,"v3 same as v2, but with short delimiters               <      $   >   >2   >="  
   ,"v1 command first; exit status is required              (none) <<< >>> >>>2 >>>="
   ,""
   ]
 
 data Args = Args {
      list        :: Bool
+    ,include     :: [String]
+    ,exclude     :: [String]
     ,all_        :: Bool
     ,color       :: Bool
     ,diff        :: Bool
     ,precise     :: Bool
     ,hide_successes :: Bool
+    ,fail_fast   :: Bool
     ,xmlout      :: String
     ,defmacro    :: [String]
-    ,include     :: [String]
-    ,exclude     :: [String]
     ,execdir     :: Bool
     ,extension   :: String
     ,with        :: String
     ,timeout     :: Int
     ,threads     :: Int
+    ,shell_cmd   :: Maybe FilePath
     ,debug       :: Bool
     ,debug_parse :: Bool
     ,testpaths   :: [FilePath]
+    ,print_      :: Maybe String
+    ,hspec_      :: Bool
     } deriving (Show, Data, Typeable)
 
 argdefs = Args {
-     list        = def     &= help "List all parsed tests and stop"
-    ,all_        = def     &= help "Don't truncate output, even if large"
-    ,color       = def     &= help "Show colored output if your terminal supports it"
-    ,diff        = def     &= name "d" &= help "Show expected output mismatches in diff format"
-    ,precise     = def     &= help "Show expected/actual output precisely (eg whitespace)"
-    ,hide_successes = def  &= help "Show only test failures"
-    ,xmlout      = def     &= typ "FILE" &= help "Specify file to store test results in xml format."
-    ,defmacro    = def  &= name "D" &= typ "D=DEF" &= help "Specify a macro that is evaluated by preprocessor before the test files are parsed. D stands for macro definition that is replaced with the value of DEF."
-    ,include     = def     &= name "i" &= typ "PAT" &= help "Include tests whose name contains this glob pattern"
+     list        = def     &= help "List the names of all tests found"
+    ,include     = def     &= name "i" &= typ "PAT" &= help "Include tests whose name contains this glob pattern (eg: -i1 -i{4,5,6})"
     ,exclude     = def     &= name "x" &= typ "STR" &= help "Exclude test files whose path contains STR"
-    ,execdir     = def     &= help "Run tests from within the test file's directory"
+    ,all_        = def     &= help "Show all output without truncating, even if large"
+    ,color       = def     &= help "Show colored output if your terminal supports it"
+    ,diff        = def     &= name "d" &= help "Show differences between expected/actual output"
+    ,precise     = def     &= name "p" &= help "Show expected/actual output precisely, with quoting"
+    ,hide_successes = def  &= explicit &= name "hide-successes" &= help "Show only test failures"
+    ,fail_fast   = def     &= explicit &= name "fail-fast" &= help "Only hspec: stop tests on first failure"
+    ,xmlout      = def     &= typ "FILE" &= help "Save test results to FILE in XML format."
+    ,defmacro    = def  &= name "D" &= typ "D=DEF" &= help "Define a macro D to be replaced by DEF while parsing test files."
+    ,execdir     = def     &= help "Run tests from within each test file's directory"
     ,extension   = ".test" &= typ "EXT" &= help "File suffix of test files (default: .test)"
-    ,with        = def     &= typ "EXECUTABLE" &= help "Replace the first word of (unindented) test commands"
+    ,with        = def     &= typ "EXE" &= help "Replace the first word of test commands with EXE (unindented commands only)"
     ,timeout     = def     &= name "o" &= typ "SECS" &= help "Number of seconds a test may run (default: no limit)"
     ,threads     = def     &= name "j" &= typ "N" &= help "Number of threads for running tests (default: 1)"
-    ,debug       = def     &= help "Show debug info, for troubleshooting"
-    ,debug_parse = def     &= help "Show test file parsing info and stop"
+    ,shell_cmd   = def     &= explicit &= name "shell" &= typ "EXE" &= help "The shell program to use (must accept -c CMD; default: /bin/sh on POSIX, cmd.exe on Windows)"
+    ,debug       = def     &= help "Show debug info while running"
+    ,debug_parse = def     &= help "Show test file parsing results and stop"
     ,testpaths   = def     &= args &= typ "TESTFILES|TESTDIRS"
+    ,print_      = def     &= typ "FORMAT" &= opt "v3" &= groupname "Print test file" &= help "Print test files in specified format (default: v3)."
+    ,hspec_      = def     &= name"hspec" &= help "Use hspec to run tests."
     }
+    &= helpArg [explicit, name "help", name "h"]
     &= program progname
     &= summary progversion
     &= details proghelpsuffix
@@ -107,6 +120,12 @@ main = do
                ++ (if timeout args > 0 then ["--timeout=" ++ show (timeout args)] else [])
                ++ (if threads args > 0 then ["--threads=" ++ show (threads args)] else [])
                ++ (if not (xmlout args == []) then ["--jxml=" ++ (xmlout args)] else [])
+      hspecconf = defaultConfig
+                  {configConcurrentJobs = Just (threads args)
+                  ,configFailFast = fail_fast args
+                  -- TODO add other options; compare shelltest -h and http://hspec.github.io/options.html
+                  -- https://hackage.haskell.org/package/hspec-core-2.7.2/docs/Test-Hspec-Core-Runner.html#g:2
+                  }
 
 
   when (debug args) $ printf "%s\n" progversion >> printf "args: %s\n" (ppShow args)
@@ -127,12 +146,32 @@ main = do
 
   -- parse test files
   when (debug args) $ printf "processing %d test files: %s\n" (length testfiles) (intercalate ", " testfiles)
-  parseresults <- mapM (parseShellTestFile (debug args || debug_parse args) preprocessor) testfiles
+  shelltests <- mapM (parseShellTestFile (debug args || debug_parse args) preprocessor) testfiles
 
-  -- run tests
-  when (debug args) $ printf "running tests:\n"
-  unless (debug_parse args) $
-    defaultMainWithArgs (concatMap (hUnitTestToTests . testFileParseToHUnitTest args) parseresults) tfopts
+  -- exit after --debug-parse output
+  if debug_parse args then exitSuccess
+  -- or print tests
+  else if isJust $ print_ args then mapM_ (printShellTestsWithResults args) shelltests
+  -- or run tests
+  else do
+    when (debug args) $ printf "running tests:\n"
+    if hspec_ args
+      then runWithHspec         args hspecconf shelltests
+      else runWithTestFramework args tfopts    shelltests
+
+runWithTestFramework :: Args -> [String] -> [Either ParseError [ShellTest]] -> IO ()
+runWithTestFramework args tfopts shelltests = defaultMainWithArgs hunittests tfopts
+  where hunittests = concatMap (hUnitTestToTests . testFileParseToHUnitTest args) shelltests
+
+runWithHspec :: Args -> Config -> [Either ParseError [ShellTest]] -> IO ()
+runWithHspec args hsconf shelltests = void $ runSpec spec hsconf
+  where
+    spec = fromHUnitTest $ TestLabel "All shelltests" $ TestList hunittests
+    hunittests = map (testFileParseToHUnitTest args) shelltests
+
+printShellTestsWithResults :: Args -> Either ParseError [ShellTest] -> IO ()
+printShellTestsWithResults args (Right ts) = mapM_ (prepareShellTest args) ts
+printShellTestsWithResults _ (Left e) = putStrLn $ "*** parse error in " ++ (sourceName $ errorPos e)
 
 -- | Additional argument checking.
 checkArgs :: Args -> IO Args
@@ -141,21 +180,18 @@ checkArgs args = do
        warn $ printf "Please specify at least one test file or directory, eg: %s tests" progname
   return args
 
--- | Show a message, usage string, and terminate with exit status 1.
-warn :: String -> IO ()
-warn s = putStrLn s >> exitWith (ExitFailure 1)
-
-
 -- running tests
 
 testFileParseToHUnitTest :: Args -> Either ParseError [ShellTest] -> Test.HUnit.Test
-testFileParseToHUnitTest args (Right ts) = TestList $ map (shellTestToHUnitTest args) ts
+testFileParseToHUnitTest args (Right ts) = TestList $ map (\t -> testname t ~: prepareShellTest args t) ts
 testFileParseToHUnitTest _ (Left e) = ("parse error in " ++ (sourceName $ errorPos e)) ~: (assertFailure :: (String -> IO ())) $ show e
 
-shellTestToHUnitTest :: Args -> ShellTest -> Test.HUnit.Test
-shellTestToHUnitTest args ShellTest{testname=n,command=c,stdin=i,stdoutExpected=o_expected,
-                                    stderrExpected=e_expected,exitCodeExpected=x_expected} =
- n ~: do
+-- | Convert this test to an IO action that either runs the test or prints it
+-- to stdout, depending on the arguments.
+prepareShellTest :: Args -> ShellTest -> IO ()
+prepareShellTest args st@ShellTest{testname=n,command=c,stdin=i,stdoutExpected=o_expected,
+                       stderrExpected=e_expected,exitCodeExpected=x_expected,lineNumber=ln} =
+ do
   let e = with args
       cmd = case (e,c) of (_:_, ReplaceableCommand s) -> e ++ " " ++ dropWhile (/=' ') s
                           (_, ReplaceableCommand s)   -> s
@@ -164,7 +200,7 @@ shellTestToHUnitTest args ShellTest{testname=n,command=c,stdin=i,stdoutExpected=
       trim' = if all_ args then id else trim
   when (debug args) $ do
     printf "actual command was: %s\n" (show cmd)
-  (o_actual, e_actual, x_actual) <- runCommandWithInput dir cmd i
+  (o_actual, e_actual, x_actual) <- runCommandWithInput (shell_cmd args) dir cmd i
   when (debug args) $ do
     printf "actual stdout was : %s\n" (show $ trim' o_actual)
     printf "actual stderr was : %s\n" (show $ trim' e_actual)
@@ -172,11 +208,13 @@ shellTestToHUnitTest args ShellTest{testname=n,command=c,stdin=i,stdoutExpected=
   let outputMatch = maybe True (o_actual `matches`) o_expected
   let errorMatch = maybe True (e_actual `matches`) e_expected
   let exitCodeMatch = show x_actual `matches` x_expected
-  if (x_actual == 127) -- catch bad executable - should work on posix systems at least
-   then ioError $ userError $ unwords $ filter (not . null) [e_actual, printf "Command: '%s' Exit code: %i" cmd x_actual] -- XXX still a test failure; should be an error
-   else assertString $ concat $ filter (not . null) [
+  case print_ args of
+    Just format -> printShellTest format st
+    Nothing -> if (x_actual == 127) -- catch bad executable - should work on posix systems at least
+           then ioError $ userError $ unwords $ filter (not . null) [e_actual, printf "Command: '%s' Exit code: %i" cmd x_actual] -- XXX still a test failure; should be an error
+           else assertString $ concat $ filter (not . null) [
              if any not [outputMatch, errorMatch, exitCodeMatch]
-               then printf "Command:\n%s\n" cmd
+               then printf "Command (at line %s):\n%s\n" (show ln) cmd
                else ""
             ,if outputMatch
               then ""
@@ -192,10 +230,10 @@ shellTestToHUnitTest args ShellTest{testname=n,command=c,stdin=i,stdoutExpected=
 -- | Run a shell command line, passing it standard input if provided,
 -- and return the standard output, standard error output and exit code.
 -- Note on unix, at least with ghc 6.12, command (and filepath) are assumed to be utf8-encoded.
-runCommandWithInput :: Maybe FilePath -> String -> Maybe String -> IO (String, String, Int)
-runCommandWithInput wd cmd input = do
+runCommandWithInput :: Maybe FilePath -> Maybe FilePath -> String -> Maybe String -> IO (String, String, Int)
+runCommandWithInput sh wd cmd input = do
   -- this has to be done carefully
-  (ih,oh,eh,ph) <- runInteractiveCommandInDir wd cmd
+  (ih,oh,eh,ph) <- runInteractiveCommandInDir sh wd cmd
   when (isJust input) $ forkIO (hPutStr ih $ fromJust input) >> return ()
   o <- newEmptyMVar
   e <- newEmptyMVar
@@ -206,22 +244,25 @@ runCommandWithInput wd cmd input = do
   e_actual <- takeMVar e
   return (o_actual, e_actual, x_actual)
 
-runInteractiveCommandInDir :: Maybe FilePath -> String ->  IO (Handle, Handle, Handle, ProcessHandle)
-runInteractiveCommandInDir wd cmd = do
+runInteractiveCommandInDir :: Maybe FilePath -> Maybe FilePath -> String ->  IO (Handle, Handle, Handle, ProcessHandle)
+runInteractiveCommandInDir sh wd cmd = do
    (mb_in, mb_out, mb_err, p) <-
       createProcess $
-         (shell cmd) { cwd = wd
-                     , std_in  = CreatePipe
-                     , std_out = CreatePipe
-                     , std_err = CreatePipe }
+         run { cwd = wd
+             , std_in  = CreatePipe
+             , std_out = CreatePipe
+             , std_err = CreatePipe }
+   -- these should all be Just since we used CreatePipe
    return (fromJust mb_in, fromJust mb_out, fromJust mb_err, p)
+   where
+      run = maybe (shell cmd) (\shcmd -> proc shcmd ["-c", cmd]) sh
 
 hGetContentsStrictlyAnd :: Handle -> (String -> IO b) -> IO b
 hGetContentsStrictlyAnd h f = hGetContents h >>= \s -> length s `seq` f s
 
 matches :: String -> Matcher -> Bool
-matches s (PositiveRegex r)   = s `containsRegex` r
-matches s (NegativeRegex r)   = not $ s `containsRegex` r
+matches s (PositiveRegex (MR r))   = s `containsRegex` r
+matches s (NegativeRegex (MR r))   = not $ s `containsRegex` r
 matches s (Numeric p)         = s == p
 matches s (NegativeNumeric p) = not $ s == p
 matches s (Lines _ p)         = s == p
